import cv2
from tensorflow.keras import utils
from tensorflow.python.keras import backend as K
from .config import BASIC_URL, FNAME_CLASSES_IN2012


def load_imgnet_weights(keras_model, weights_npz):
    """Load the imagenet pre training weights (.npz) for ViT Keras model.
    Weights file, for example: "ViT-B_16_imagenet21k.npz", "ViT-L_32_imagenet21k+imagenet2012.npz"
    More weights files: https://github.com/djsaber/Keras-ViT/releases/tag/v1.0.0
    args:
        - keras_model: The keras ViT models that need to load pre trained weights
        - weights_npz: The .npz weights file
    return:
        - load_inf: Dict of information about loading weights for each layer of the keras_model
    """
    params_dict = K.np.load(weights_npz, allow_pickle=False)
    keys = set(params_dict.keys())
    transformer_blocks_num = len([k for k in keys if k.startswith("Transformer/encoderblock_")])//16
    
    patch_emb_weights = [
        params_dict["embedding/kernel"],
        params_dict["embedding/bias"]]
    cls_token_weights = [params_dict["cls"]]
    position_emb_weights = [params_dict["Transformer/posembed_input/pos_embedding"]]
    transformer_block_weights = [
        [params_dict[f"Transformer/encoderblock_{i}/MultiHeadDotProductAttention_1/query/kernel"], 
         params_dict[f"Transformer/encoderblock_{i}/MultiHeadDotProductAttention_1/query/bias"],
         params_dict[f"Transformer/encoderblock_{i}/MultiHeadDotProductAttention_1/key/kernel"], 
         params_dict[f"Transformer/encoderblock_{i}/MultiHeadDotProductAttention_1/key/bias"], 
         params_dict[f"Transformer/encoderblock_{i}/MultiHeadDotProductAttention_1/value/kernel"], 
         params_dict[f"Transformer/encoderblock_{i}/MultiHeadDotProductAttention_1/value/bias"], 
         params_dict[f"Transformer/encoderblock_{i}/MultiHeadDotProductAttention_1/out/kernel"], 
         params_dict[f"Transformer/encoderblock_{i}/MultiHeadDotProductAttention_1/out/bias"], 
         params_dict[f"Transformer/encoderblock_{i}/MlpBlock_3/Dense_0/kernel"], 
         params_dict[f"Transformer/encoderblock_{i}/MlpBlock_3/Dense_0/bias"], 
         params_dict[f"Transformer/encoderblock_{i}/MlpBlock_3/Dense_1/kernel"], 
         params_dict[f"Transformer/encoderblock_{i}/MlpBlock_3/Dense_1/bias"], 
         params_dict[f"Transformer/encoderblock_{i}/LayerNorm_0/scale"], 
         params_dict[f"Transformer/encoderblock_{i}/LayerNorm_0/bias"], 
         params_dict[f"Transformer/encoderblock_{i}/LayerNorm_2/scale"], 
         params_dict[f"Transformer/encoderblock_{i}/LayerNorm_2/bias"]
         ]  for i in range(transformer_blocks_num)]
    layer_norm_weights = [
        params_dict["Transformer/encoder_norm/scale"],
        params_dict["Transformer/encoder_norm/bias"]]
    pre_logic_weights = [
            params_dict["pre_logits/kernel"],
            params_dict["pre_logits/bias"]] if "pre_logits/kernel" in keys else []
    mlp_head_weights = [
        params_dict["head/kernel"],
        params_dict["head/bias"]]
    
    load_inf = {}
    for l in keras_model.layers:
        if l.name == "patch_embedding":
            try:
                l.set_weights(patch_emb_weights)
                load_inf[l.name] = "loaded"
            except: load_inf[l.name] = "mismatch"
    
        elif l.name == "add_cls_token":
            try:
                l.set_weights(cls_token_weights)
                load_inf[l.name] = "loaded - imagenet"
            except: load_inf[l.name] = "not loaded - mismatch"
    
        elif l.name == "position_embedding":
            try:
                l.set_weights(position_emb_weights)
                load_inf[l.name] = "loaded - imagenet"
            except: load_inf[l.name] = "not loaded - mismatch"
    
        elif l.name.startswith("transformer_block_"):
            i = int(l.name[l.name.rfind("_")+1:])
            try:
                weights = [w.reshape(s.shape) for w,s in zip(transformer_block_weights[i], l.weights)]
                l.set_weights(weights)
                load_inf[l.name] = "loaded - imagenet"
            except: load_inf[l.name] = "not loaded - mismatch"
    
        elif l.name == "layer_norm":
            try:
                l.set_weights(layer_norm_weights)
                load_inf[l.name] = "loaded - imagenet"
            except: load_inf[l.name] = "not loaded - mismatch"
    
        elif l.name == "pre_logits":
            try:
                l.set_weights(pre_logic_weights)
                load_inf[l.name] = "loaded - imagenet"
            except: 
                if pre_logic_weights == []:
                    load_inf[l.name] = "not loaded - not found"
                else:
                    load_inf[l.name] = "not loaded - mismatch"
    
        elif l.name == "mlp_head":
            try:
                l.set_weights(mlp_head_weights)
                load_inf[l.name] = "loaded - imagenet"
            except: load_inf[l.name] = "not loaded - mismatch"
    
    keras_model.load_weights_inf = load_inf


def get_imagenet2012_classes():
    """Get the dict of ImageNet 2012 classes."""
    classes_txt = utils.get_file(FNAME_CLASSES_IN2012, origin=BASIC_URL+FNAME_CLASSES_IN2012, cache_subdir="classes")
    with open(classes_txt, encoding="utf-8") as f:
        classes = {i:l.strip() for i,l in enumerate(f.readlines())}
    return classes


def read_img(file_path, resize=None):
    """Read an image and scale it to [-1,1].
    args:
        - file_path: The image file path
        - resize: The target size of image
    return:
        - An np array with values between 0 and 1
    
    """
    img = cv2.imread(file_path)
    img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
    if resize:
        img = cv2.resize(img, resize)
    return img/127.5-1


BASIC_URL = "https://github.com/djsaber/Keras-ViT/releases/download/v1.0.0/"

SIZE = ["ViT-B16", "ViT-B32", "ViT-L16", "ViT-L32"]

WEIGHTS = ["imagenet21k", "imagenet21k+imagenet2012"]

FNAMES = [
    "ViT-B_16_imagenet21k.npz", 
    "ViT-B_32_imagenet21k.npz",
    "ViT-L_16_imagenet21k.npz",
    "ViT-L_32_imagenet21k.npz",
    "ViT-B_16_imagenet21k+imagenet2012.npz",
    "ViT-B_32_imagenet21k+imagenet2012.npz",
    "ViT-L_16_imagenet21k+imagenet2012.npz",
    "ViT-L_32_imagenet21k+imagenet2012.npz"]

PRE_TRAINED_WEIGHTS_FNAME = {
    SIZE[0]: {WEIGHTS[0]: FNAMES[0], WEIGHTS[1]: FNAMES[4]},
    SIZE[1]: {WEIGHTS[0]: FNAMES[1], WEIGHTS[1]: FNAMES[5]},
    SIZE[2]: {WEIGHTS[0]: FNAMES[2], WEIGHTS[1]: FNAMES[6]},
    SIZE[3]: {WEIGHTS[0]: FNAMES[3], WEIGHTS[1]: FNAMES[7]},
    }

WEIGHTS_CONFIG = {
    WEIGHTS[0]: {"IMAGE_SIZE": 224, "NUM_CLASSES": 21843, "PRE_LOGITS": True},
    WEIGHTS[1]: {"IMAGE_SIZE": 384, "NUM_CLASSES": 1000, "PRE_LOGITS": False}
    }

VIT_B_CONFIG = {
    "hidden_dim": 768,
    "mlp_dim": 3072,
    "atten_heads": 12,
    "encoder_depth": 12,
    }

VIT_L_CONFIG = {
    "hidden_dim": 1024,
    "mlp_dim": 4096,
    "atten_heads": 16,
    "encoder_depth": 24,
    }

FNAME_CLASSES_IN2012 = "imagenet2012.txt"


from tensorflow.keras import utils
from tensorflow.keras import models
from tensorflow.keras import layers
from tensorflow.keras import activations
from .utils import load_imgnet_weights
from .layers import PatchEmbedding, AddCLSToken, AddPositionEmbedding, TransformerEncoder
from .config import BASIC_URL, VIT_B_CONFIG, VIT_L_CONFIG, WEIGHTS_CONFIG, PRE_TRAINED_WEIGHTS_FNAME




class ViT(models.Model):
    """Implementation of VisionTransformer Based on Keras
    Args:
        - image_size: Input Image Size, integer or tuple
        - patch_size: Size of each patch, integer or tuple
        - num_classes: Number of output classes
        - hidden_dim: The embedding dimension of each patch, it should be set to an integer multiple of the `atten_heads`
        - mlp_dim: The projection dimension of the mlp_block, it is generally 4 times that of `hidden_dim`
        - atten_heads: Number of self attention heads
        - encoder_depth: Number of the transformer encoder layer
        - dropout_rate: Dropout probability
        - activatiion: Activation function of mlp_head
        - pre_logits: Insert pre_Logits layer or not 
        - include_mlp_head: Insert mlp_head layer or not
    """
    def __init__(
        self, 
        image_size = 224,
        patch_size = 16,
        num_classes = 1000,
        hidden_dim = 768,
        mlp_dim = 3072,
        atten_heads = 12,
        encoder_depth = 12,
        dropout_rate = 0.,
        activation = "linear",
        pre_logits = True,
        include_mlp_head = True,
        **kwargs
        ):
        assert isinstance(image_size, int) or isinstance(image_size, tuple), "`image_size` should be int or tuple !"
        assert isinstance(patch_size, int) or isinstance(patch_size, tuple), "`patch_size` should be int or tuple !"
        assert hidden_dim%atten_heads==0, "hidden_dim and atten_heads do not match !"

        if isinstance(image_size, int):
            image_size = (image_size, image_size)
        if isinstance(patch_size, int):
            patch_size = (patch_size, patch_size)

        if not kwargs.get("name"):
            config = {
                "hidden_dim": hidden_dim,
                "mlp_dim": mlp_dim,
                "atten_heads": atten_heads,
                "encoder_depth": encoder_depth,
                }
            size = "CUSTOM_SIZE"
            if config == VIT_B_CONFIG: size = "B"
            if config == VIT_L_CONFIG: size = "L" 
            patch = f"{patch_size[0]}" if patch_size[0]==patch_size[1] else f"{patch_size[0]}_{patch_size[1]}"
            image = f"{image_size[0]}" if image_size[0]==image_size[1] else f"{image_size[0]}_{image_size[1]}"
            kwargs["name"] = f"ViT-{size}-{patch}-{image}"

        super().__init__(**kwargs)
        self.image_size = image_size[:2]
        self.patch_size = patch_size[:2]
        self.hidden_dim = hidden_dim
        self.mlp_dim = mlp_dim
        self.num_classes = num_classes
        self.atten_heads = atten_heads
        self.encoder_depth = encoder_depth
        self.dropout_rate = dropout_rate
        self.activation = activations.get(activation)
        self.pre_logits = pre_logits
        self.include_mlp_head = include_mlp_head
        self.load_weights_inf = None
        self.build((None, *self.image_size, 3))

    def build(self, input_shape):
        self.patch_embedding=PatchEmbedding(self.patch_size, self.hidden_dim, name="patch_embedding")
        self.add_cls_token = AddCLSToken(self.hidden_dim, name="add_cls_token")
        self.position_embedding = AddPositionEmbedding(name="position_embedding")
        self.encoder_blocks = [
            TransformerEncoder(
                self.mlp_dim, 
                self.atten_heads, 
                self.dropout_rate,
                name=f"transformer_block_{i}"
                ) for i in range(self.encoder_depth)]
        self.layer_norm = layers.LayerNormalization(epsilon=1e-6, name="layer_norm")
        self.extract_token = layers.Lambda(lambda x: x[:,0], name="extract_token")
        if self.pre_logits: self.pre_logits = layers.Dense(self.hidden_dim, activation="tanh", name="pre_logits")
        if self.include_mlp_head: self.mlp_head = layers.Dense(self.num_classes, self.activation, name="mlp_head")
        super().build(input_shape)
        self.call(layers.Input(input_shape[1:]))

    def call(self, inputs):
        x = self.patch_embedding(inputs)
        x = self.add_cls_token(x)
        x = self.position_embedding(x)
        for encoder in self.encoder_blocks:
            x = encoder(x)
        x = self.layer_norm(x)
        x = self.extract_token(x)
        if self.pre_logits:
            x = self.pre_logits(x)
        if self.include_mlp_head:
            x = self.mlp_head(x)
        return x

    def load_weights(self, *kwargs):
        super().load_weights(*kwargs)
        self.load_weights_inf = {l.name:"loaded - h5 weights" for l in self.layers}

    def loading_summary(self):
        """ Print model information about loading imagenet pre training weights.
        ` - imagenet` means successful reading of imagenet pre training weights file
        ` - not pre trained` means that the model did not load the imagenet pre training weights file
        ` - mismatch` means that the pre training weights of the imagenet do not match the parameter shapes of the model layer
        ` - not found` means that there is no weights for the layer in the imagenet pre training weights file
        ` - h5 weights` means that the model is loaded with h5 weights
        """
        if self.load_weights_inf:
            inf = [(key+' '*(35-len(key))+value+"\n") for (key,value) in self.load_weights_inf.items()]
        else:
            inf = [(l.name+' '*(35-len(l.name))+"not loaded - not pre trained"+"\n") for l in self.layers]
        print(f'Model: "{self.name}"')
        print("-"*65)
        print("layers" + " "*29 + "load weights inf")
        print("="*65)
        print("\n".join(inf)[:-1])
        print("="*65)


def ViT_B16(
    image_size=None,
    num_classes=None,
    activation="linear",
    dropout_rate=0.1,
    pre_logits=None,
    include_mlp_head=True,
    pre_trained=True,
    weights="imagenet21k",
    **kwargs
    ):
    """Implementation of ViT_B16 Based on Keras.
    Args:
        - image_size: Input Image Size
        - num_classes: Number of output classes
        - activation: Activation function of mlp_head
        - dropout_rate: Dropout probability
        - pre_logits: Insert pre_Logits layer or not 
        - include_mlp_head: Insert mlp_head layer or not
        - pre_trained: using pre trained model weights or not
        the .npz weight file will be downloaded to "C:\\Users\\username\\.keras\\weights" when it is True
        - weights: This argument will determine the selection which dataset does the pre trained model come from
        this argument will set the values of argument `image_size`, `num_classes`, and `pre_logits` if they are None
        this argument should be set to one of "imagenet21k" and "imagenet21k+imagenet2012"
        "imagenet21k" >>> `image_size`=224, `num_classes`=21843, `pre_logits`=True
        "imagenet21k+imagenet2012" >>> `image_size`=384, `num_classes`=1000, `pre_logits`=False

    return:
        - ViT_B16: ViT with a patch_size of 16, hiddem_dim of 768, 
        mlp_dim of 3072, atten_heads of 12, and encoder layers of 12
    """
 
    if image_size==None: image_size=WEIGHTS_CONFIG.get(weights).get("IMAGE_SIZE")
    if num_classes==None: num_classes=WEIGHTS_CONFIG.get(weights).get("NUM_CLASSES")
    if pre_logits==None: pre_logits=WEIGHTS_CONFIG.get(weights).get("PRE_LOGITS")

    vit = ViT(
        **VIT_B_CONFIG,
        patch_size=16,
        image_size=image_size,
        num_classes=num_classes,
        activation=activation,
        dropout_rate=dropout_rate,
        pre_logits=pre_logits,
        include_mlp_head=include_mlp_head,
        **kwargs
        )

    if pre_trained:
        FNAME = PRE_TRAINED_WEIGHTS_FNAME["ViT-B16"][weights]
        weights = utils.get_file(FNAME, origin=BASIC_URL+FNAME, cache_subdir="weights")
        load_imgnet_weights(vit, weights)

    return vit


def ViT_B32(
    image_size=None,
    num_classes=None,
    activation="linear",
    dropout_rate=0.1,
    pre_logits=None,
    include_mlp_head=True,
    pre_trained=True,
    weights="imagenet21k",
    **kwargs
    ):
    """Implementation of ViT_B32 Based on Keras.
    Args:
        - image_size: Input Image Size
        - num_classes: Number of output classes
        - activation: Activation function of mlp_head
        - dropout_rate: Dropout probability
        - pre_logits: Insert pre_Logits layer or not 
        - include_mlp_head: Insert mlp_head layer or not
        - pre_trained: using pre trained model weights or not
        the .npz weight file will be downloaded to "C:\\Users\\username\\.keras\\weights" when it is True
        - weights: This argument will determine the selection which dataset does the pre trained model come from
        this argument will set the values of argument `image_size`, `num_classes`, and `pre_logits` if they are None
        this argument should be set to one of "imagenet21k" and "imagenet21k+imagenet2012"
        "imagenet21k" >>> `image_size`=224, `num_classes`=21843, `pre_logits`=True
        "imagenet21k+imagenet2012" >>> `image_size`=384, `num_classes`=1000, `pre_logits`=False

    return:
        - ViT_B32: ViT with a patch_size of 32, hiddem_dim of 768, 
        mlp_dim of 3072, atten_heads of 12, and encoder layers of 12
    """

    if image_size==None: image_size=WEIGHTS_CONFIG.get(weights).get("IMAGE_SIZE")
    if num_classes==None: num_classes=WEIGHTS_CONFIG.get(weights).get("NUM_CLASSES")
    if pre_logits==None: pre_logits=WEIGHTS_CONFIG.get(weights).get("PRE_LOGITS")

    vit = ViT(
        **VIT_B_CONFIG,
        patch_size=32,
        image_size=image_size,
        num_classes=num_classes,
        activation=activation,
        dropout_rate=dropout_rate,
        pre_logits=pre_logits,
        include_mlp_head=include_mlp_head,
        **kwargs,
        )

    if pre_trained:
        FNAME = PRE_TRAINED_WEIGHTS_FNAME["ViT-B32"][weights]
        weights = utils.get_file(FNAME, origin=BASIC_URL+FNAME, cache_subdir="weights")
        load_imgnet_weights(vit, weights)

    return vit


def ViT_L16(
    image_size=None,
    num_classes=None,
    activation="linear",
    dropout_rate=0.1,
    pre_logits=None,
    include_mlp_head=True,
    pre_trained=True,
    weights="imagenet21k",
    **kwargs
    ):
    """Implementation of ViT_L16 Based on Keras.
    Args:
        - image_size: Input Image Size
        - num_classes: Number of output classes
        - activation: Activation function of mlp_head
        - dropout_rate: Dropout probability
        - pre_logits: Insert pre_Logits layer or not 
        - include_mlp_head: Insert mlp_head layer or not
        - pre_trained: using pre trained model weights or not
        the .npz weight file will be downloaded to "C:\\Users\\username\\.keras\\weights" when it is True
        - weights: This argument will determine the selection which dataset does the pre trained model come from
        this argument will set the values of argument `image_size`, `num_classes`, and `pre_logits` if they are None
        this argument should be set to one of "imagenet21k" and "imagenet21k+imagenet2012"
        "imagenet21k" >>> `image_size`=224, `num_classes`=21843, `pre_logits`=True
        "imagenet21k+imagenet2012" >>> `image_size`=384, `num_classes`=1000, `pre_logits`=False

    return:
        - ViT_L16: ViT with a patch_size of 16, hiddem_dim of 1024, 
        mlp_dim of 4096, atten_heads of 16, and encoder layers of 24
    """

    if image_size==None: image_size=WEIGHTS_CONFIG.get(weights).get("IMAGE_SIZE")
    if num_classes==None: num_classes=WEIGHTS_CONFIG.get(weights).get("NUM_CLASSES")
    if pre_logits==None: pre_logits=WEIGHTS_CONFIG.get(weights).get("PRE_LOGITS")

    vit = ViT(
        **VIT_L_CONFIG,
        patch_size=16,
        image_size=image_size,
        num_classes=num_classes,
        activation=activation,
        dropout_rate=dropout_rate,
        pre_logits=pre_logits,
        include_mlp_head=include_mlp_head,
        **kwargs,
        )

    if pre_trained:
        FNAME = PRE_TRAINED_WEIGHTS_FNAME["ViT-L16"][weights]
        weights = utils.get_file(FNAME, origin=BASIC_URL+FNAME, cache_subdir="weights")
        load_imgnet_weights(vit, weights)

    return vit


def ViT_L32(
    image_size=None,
    num_classes=None,
    activation="linear",
    dropout_rate=0.1,
    pre_logits=None,
    include_mlp_head=True,
    pre_trained=True,
    weights="imagenet21k",
    **kwargs
    ):
    """Implementation of ViT_L32 Based on Keras.
    Args:
        - image_size: Input Image Size
        - num_classes: Number of output classes
        - activation: Activation function of mlp_head
        - dropout_rate: Dropout probability
        - pre_logits: Insert pre_Logits layer or not 
        - include_mlp_head: Insert mlp_head layer or not
        - pre_trained: using pre trained model weights or not
        the .npz weight file will be downloaded to "C:\\Users\\username\\.keras\\weights" when it is True
        - weights: This argument will determine the selection which dataset does the pre trained model come from
        this argument will set the values of argument `image_size`, `num_classes`, and `pre_logits` if they are None
        this argument should be set to one of "imagenet21k" and "imagenet21k+imagenet2012"
        "imagenet21k" >>> `image_size`=224, `num_classes`=21843, `pre_logits`=True
        "imagenet21k+imagenet2012" >>> `image_size`=384, `num_classes`=1000, `pre_logits`=False

    return:
        - ViT_L16: ViT with a patch_size of 32, hiddem_dim of 1024, 
        mlp_dim of 4096, atten_heads of 16, and encoder layers of 24
    """

    if image_size==None: image_size=WEIGHTS_CONFIG.get(weights).get("IMAGE_SIZE")
    if num_classes==None: num_classes=WEIGHTS_CONFIG.get(weights).get("NUM_CLASSES")
    if pre_logits==None: pre_logits=WEIGHTS_CONFIG.get(weights).get("PRE_LOGITS")

    vit = ViT(
        **VIT_L_CONFIG,
        patch_size=32,
        image_size=image_size,
        num_classes=num_classes,
        activation=activation,
        dropout_rate=dropout_rate,
        pre_logits=pre_logits,
        include_mlp_head=include_mlp_head,
        **kwargs,
        )

    if pre_trained:
        FNAME = PRE_TRAINED_WEIGHTS_FNAME["ViT-L32"][weights]
        weights = utils.get_file(FNAME, origin=BASIC_URL+FNAME, cache_subdir="weights")
        load_imgnet_weights(vit, weights)

    return vit


import setuptools
 
with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()
 
setuptools.setup(
    name="vit-tf2",
    version="1.0.1",
    author="djsaber",
    author_email="479719615@qq.com",
    description="Implementation of ViT model based on tensorflow",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/djsaber/Keras-ViT",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
)



import tensorflow as tf
from tensorflow.keras import layers
from tensorflow.keras import activations


class PatchEmbedding(layers.Layer):
    """patch embedding layer"""
    def __init__(self, patch_size, hidden_dim, **kwargs):
        super().__init__(**kwargs)
        self.patch_size = patch_size
        self.hidden_dim = hidden_dim

    def build(self, input_shape):
        self.conv = layers.Conv2D(self.hidden_dim, self.patch_size, self.patch_size)
        self.flatten = layers.Reshape((-1, self.hidden_dim))
        super().build(input_shape)

    def call(self, inputs):
        x = self.conv(inputs)
        x = self.flatten(x)
        return x


class AddCLSToken(layers.Layer):
    """add class token layer"""
    def __init__(self, hidden_dim, **kwargs):
        super().__init__(**kwargs)
        self.hidden_dim = hidden_dim

    def build(self, input_shape):
        self.class_token = self.add_weight(
            name="class_token_weight",
            shape=(1, 1, self.hidden_dim),
            initializer="zero",
            trainable=True
            )
        return super().build(input_shape)

    def call(self, inputs):   
        x = tf.tile(self.class_token, [tf.shape(inputs)[0],1,1])
        x = tf.concat([x, inputs], axis=1)
        return x


class AddPositionEmbedding(layers.Layer):
    """add position encoding layer"""
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def build(self, input_shape):
        self.pe = self.add_weight(
            name="position_embedding",
            shape=(1, input_shape[1], input_shape[2]),
            initializer='random_normal',
            trainable=True
            )
        return super().build(input_shape)

    def call(self, inputs):
        return inputs + self.pe


class MultiHeadAttention(layers.Layer):
    """multi head self attention layer"""
    def __init__(
        self, 
        heads,
        **kwargs
        ):
        super().__init__(**kwargs)
        self.heads = heads

    def build(self, input_shape):
        self.dk = tf.sqrt(tf.cast(input_shape[-1]//self.heads, dtype=tf.float32))
        self.q_dense = layers.Dense(input_shape[-1], name="query")
        self.k_dense = layers.Dense(input_shape[-1], name="key")
        self.v_dense = layers.Dense(input_shape[-1], name="value")
        self.o_dense = layers.Dense(input_shape[-1], name="combine_out")
        return super().build(input_shape)

    def call(self, inputs):
        q = self.q_dense(inputs)
        k = self.k_dense(inputs)
        v = self.v_dense(inputs)
        q = tf.concat(tf.split(q, self.heads, axis=-1), axis=0)
        k = tf.concat(tf.split(k, self.heads, axis=-1), axis=0)
        v = tf.concat(tf.split(v, self.heads, axis=-1), axis=0)
        qk = tf.matmul(q, k, transpose_b=True)  
        qk = activations.softmax(qk / self.dk)
        qkv = tf.matmul(qk, v)
        qkv = tf.concat(tf.split(qkv, self.heads, axis=0), axis=-1)
        return self.o_dense(qkv)


class MLPBlock(layers.Layer):
    """mlp block"""
    def __init__(self, liner_dim, dropout_rate, **kwargs):
        self.liner_dim = liner_dim
        self.dropout_rate = dropout_rate
        super().__init__(**kwargs)

    def build(self, input_shape):
        self.dropout = layers.Dropout(self.dropout_rate)
        self.liner_1 = layers.Dense(self.liner_dim, activations.gelu)
        self.liner_2 = layers.Dense(input_shape[-1])
        return super().build(input_shape)

    def call(self, inputs):
        h = self.liner_1(inputs)
        h = self.dropout(h)
        h = self.liner_2(h)
        h = self.dropout(h)
        return h


class TransformerEncoder(layers.Layer):
    """transformerd encoder block"""
    def __init__(
        self, 
        liner_dim,
        atten_heads,
        dropout_rate,
        **kwargs
        ):
        super().__init__(**kwargs)
        self.liner_dim = liner_dim
        self.atten_heads = atten_heads
        self.dropout_rate = dropout_rate

    def build(self, input_shape):    
        self.multi_head_attens = MultiHeadAttention(
            name='multi_head_attention_layer',
            heads=self.atten_heads, 
            )
        self.mlp_block = MLPBlock(
            name='mlp_block_layer',
            liner_dim=self.liner_dim,
            dropout_rate = self.dropout_rate
            )
        self.layer_norm_1 = layers.LayerNormalization(epsilon=1e-6)
        self.layer_norm_2 = layers.LayerNormalization(epsilon=1e-6)
        self.dropout = layers.Dropout(self.dropout_rate)
        super().build(input_shape)

    def call(self, inputs):
        x = self.layer_norm_1(inputs)
        x = self.multi_head_attens(x)
        x = self.dropout(x)
        x = x + inputs
        y = self.layer_norm_2(x)
        y = self.mlp_block(y)
        return x + y